/*
 * Created on 30-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package hipi.compiler;

import java.io.File;
import java.io.FileNotFoundException;

/**
 * 
 * Replaces ASTImportDeclaration nodes with the tree related to 
 * the imported file.
 * 
 * @author Paolo Milazzo
 */
public class BuildImportVisitor implements HiPiCompilerVisitor {

	// the root node of the main program
	private ASTHiPiProgram rootNode;
	
	// a reference to the main parser (used to call error handling methods)
	private HiPiCompiler parser;
	
	// default constructor
	public BuildImportVisitor(HiPiCompiler p) {
		parser = p;
	}
	
	public Object visit(SimpleNode node, Object data) {
		return data;
	}

	public Object visit(ASTHiPiProgram node, Object data) {
		// substitutes Import nodes with imported tree (ignores imported global channels)
		for (int i=0; i<node.jjtGetNumChildren(); i++) {
			rootNode = node;
			i = ((Integer)node.jjtGetChild(i).jjtAccept(this, new Integer(i))).intValue();
		}
		// change child nodes ordering (because scheds and global recdefs are shuffled)
		int left = 0; int right = node.jjtGetNumChildren()-1;
		for (int i=0; i<=right; i++) {
			Node current = node.jjtGetChild(i);
			if ((ASTGlobalChannelDeclaration.class.isInstance(current)) 
				|| (ASTTypeDeclaration.class.isInstance(current))){
					left++;
			} 
			else if (ASTRecursiveDefinition.class.isInstance(current)) {
				node.removeChild(i);
				node.insertChild(current,left);
			}
			else if (ASTSchedule.class.isInstance(current)) {
				node.removeChild(i);
				node.insertChild(current,right);
			}
		}
		
		return null;
	}

	public Object visit(ASTGlobalChannelDeclaration node, Object data) {
		return data;
	}

	public Object visit(ASTTypeDeclaration node, Object data) {
		return data;
	}

	// executes the substituition
	public Object visit(ASTImportDeclaration node, Object data) {
		try {
			int index = ((Integer)data).intValue();
			// returns the working directory
			String workDir = 
				new File(HiPiCompiler.sourceFile).getParent();
			// returns the abstract name of the imported file 
			String fileName = new File(workDir,node.fileName).getAbsolutePath();
			HiPiCompiler comp = new HiPiCompiler(new java.io.FileInputStream(fileName));
			ASTHiPiProgram p = comp.HiPiProgram();
			// removes global channel declarations from imported file
			for (int j=0; j<p.jjtGetNumChildren(); j++) {
				if (ASTGlobalChannelDeclaration.class.isInstance(p.jjtGetChild(j))) {
					p.removeChild(j--);
				}
			}
			// removes the ImportDeclaration node
			rootNode.removeChild(index);
			// adds imported nodes
			for (int j=0; j<p.jjtGetNumChildren(); j++) {
				rootNode.insertChild(p.jjtGetChild(j),index+j);
			}
			return new Integer(--index);
		} catch (FileNotFoundException e) {
			parser.hipi_import_error(node.line,node.fileName + " not found");
		} catch (ParseException e) {
			parser.hipi_import_error(node.line,"Parse error on imported file");
		}
		return data;
	}

	public Object visit(ASTRecursiveDefinition node, Object data) {
		return data;
	}

	public Object visit(ASTSchedule node, Object data) {
		return data;
	}

	public Object visit(ASTScheduleMain node, Object data) {
		return data;
	}

	public Object visit(ASTBlock node, Object data) {
		return data;
	}

	public Object visit(ASTLocalDeclarationStatement node, Object data) {
		return data;
	}

	public Object visit(ASTSend node, Object data) {
		return data;
	}

	public Object visit(ASTRecv node, Object data) {
		return data;
	}

	public Object visit(ASTRecursiveDefinitionCall node, Object data) {
		return data;
	}

	public Object visit(ASTSpawnStatement node, Object data) {
		return data;
	}

	public Object visit(ASTSelectionStatement node, Object data) {
		return data;
	}

	public Object visit(ASTReturnStatement node, Object data) {
		return data;
	}

	public Object visit(ASTExpOp node, Object data) {
		return data;
	}

	public Object visit(ASTExpIdentifier node, Object data) {
		return data;
	}

	public Object visit(ASTExpInteger node, Object data) {
		return data;
	}

	public Object visit(ASTExpString node, Object data) {
		return data;
	}

	public Object visit(ASTNewChannel node, Object data) {
		return data;
	}
}
