package hipi.compiler.checker;

import hipi.compiler.util.StringMap;

import java.util.Vector;

public class SymbolTable {

    private StringMap symTab;
    private Vector envStack;
    private int currentEnvDepth;
    private SymbolTableEnvironment currentEnv;

    public SymbolTable() {
        symTab = new StringMap();
        envStack = new Vector();
        currentEnv = null;
		currentEnvDepth = 0; 
    }
    
    public void addElement(SymbolTableElement ste) throws SymbolTableException {
        
        // check the name value
        String name = ste.getName();
        SymbolTableElement previous = null;
        if (symTab.containsKey(name)) {
        	previous = (SymbolTableElement) symTab.remove(name);
        	if (previous.getDepth() == currentEnvDepth)
            	throw new SymbolTableException(name + " already declared");
        }
        
        // add the new name to the symbol table and to the current environment
        ste.setPrevious(previous);
        ste.setDepth(currentEnvDepth);
        symTab.put(name,ste);
        currentEnv = new SymbolTableEnvironment(name, ste, currentEnv);
    }
    
    public SymbolTableElement getElement(String name) throws SymbolTableException {
			 // check the name value
			 if (!symTab.containsKey(name)) {
				 throw new SymbolTableException(name + " undeclared");
			 }

			 // return the type value
			 return ((SymbolTableElement) symTab.get(name));
    	
    }
    
    public void addEnv() {
    	envStack.add(currentEnvDepth++,currentEnv);
    	currentEnv = null;
    }
        
    public void removeEnv() throws SymbolTableException {

		// for each element in current environment
		while (currentEnv != null) {       
        	
        	// check the name value
        	String name = currentEnv.name;	
        	if (!symTab.containsKey(name)) {
        	    throw new SymbolTableException(name + " undeclared");
        	}	
        	
        	// remove the name from the symbol table
        	SymbolTableElement previous = ((SymbolTableElement) symTab.remove(name)).getPrevious();
        	if (previous != null) {
        		symTab.put(name, previous);
        	}
        	currentEnv = currentEnv.nextElem;
		}
		
		//update the environment stack
		if (currentEnvDepth==0) {
			throw new SymbolTableException("wrong environment depth");			
		}
		currentEnv = (SymbolTableEnvironment) envStack.remove(--currentEnvDepth);
    }
    
    public int getType(String name) throws SymbolTableException {
        
        // check the name value
        if (!symTab.containsKey(name)) {
            throw new SymbolTableException(name + " undeclared");
        }
        
        // return the type value
        return ((SymbolTableElement) symTab.get(name)).getType();
    }
    
    public int getRealType(String name) throws SymbolTableException {
        
	// check the name value
	if (!symTab.containsKey(name)) {
		throw new SymbolTableException(name + " undeclared");
	}
        
	// return the type value
	return ((SymbolTableElement) symTab.get(name)).getRealType();
}
    
    
}