/*
 * Created on 21-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package hipi.compiler.encodings;

import java.util.HashSet;
import java.util.Vector;

import hipi.compiler.*;

/**
 * 
 * @author Paolo Milazzo
 */
public class RecDefsEncoding implements HiPiCompilerVisitor, HiPiCompilerTreeConstants {
	
	private static final int RESERVED_ENV_IDX_PARAMS = 0;
	private static final int RESERVED_ENV_IDX_RETVAL = 1;

	// this is used to prevent the visit of nodes added by
	// this encoding.
	// TODO better solution
	private HashSet addedNodes = new HashSet();

	public Object visit(SimpleNode node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTHiPiProgram node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTGlobalChannelDeclaration node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecursiveDefinition node, Object data) {
		childrenEncode(node);
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSchedule node, Object data) {
		childrenEncode(node);
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTScheduleMain node, Object data) {
		childrenEncode(node);
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTBlock node, Object data) {
		childrenEncode(node);		
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTLocalDeclarationStatement node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSend node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecv node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTRecursiveDefinitionCall node, Object data) {		
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTSpawnStatement node, Object data) {
		if (!addedNodes.contains(node)) {
			childrenEncode(node);
			node.childrenAccept(this,data);
		}
		return null;
	}

	public Object visit(ASTSelectionStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTReturnStatement node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpOp node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpIdentifier node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpInteger node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTExpString node, Object data) {
		node.childrenAccept(this,data);
		return null;
	}

	public Object visit(ASTNewChannel node, Object data) {
		if (!addedNodes.contains(node))
			node.childrenAccept(this,data);
		return null;
	}

	private void childrenEncode(SimpleNode node) {
		SimpleNode child;
		for (int i=0; i<node.jjtGetNumChildren() ; i++) {
			child = (SimpleNode) node.jjtGetChild(i);
			
			// if it is assigned to a variable
			if (ASTLocalDeclarationStatement.class.isInstance(child)) {
				Node subchild = child.jjtGetChild(0);
				if (ASTRecursiveDefinitionCall.class.isInstance(subchild)) 
					i+=encode(node,i,child,subchild, false);
			}
			// if it is called without assignment
			else if (ASTRecursiveDefinitionCall.class.isInstance(child))
				i+=encode(node,i,node,child,true);
		}
	}
	
	// executes the encoding and returns the number of added nodes 
	private int encode(SimpleNode node, int idx, Node child, Node subchild, boolean isVoid) {
	
		// TODO by now only void functions can be called w/o assignment

		/*
		 * Encoding:
		 * [y=f(x);P] = (u)(v)({body of f}|u.send(x)|v.recv(y);[P])
		 */
		
		// create new channel for parameters (u)
		ASTLocalDeclarationStatement astLocal1 = new ASTLocalDeclarationStatement(JJTLOCALDECLARATIONSTATEMENT);
		addedNodes.add(astLocal1);
		ASTNewChannel astNew1 = new ASTNewChannel(JJTNEWCHANNEL);
		addedNodes.add(astNew1);
		astLocal1.envIndex = RESERVED_ENV_IDX_PARAMS;
		astLocal1.TC_visited = true; // paranoid
		astLocal1.type = Types.CHANNEL;
		astLocal1.jjtAddChild(astNew1,0);
			
		// create new channel for return value (v)
		ASTLocalDeclarationStatement astLocal2 = new ASTLocalDeclarationStatement(JJTLOCALDECLARATIONSTATEMENT);
		addedNodes.add(astLocal2);
		ASTNewChannel astNew2 = new ASTNewChannel(JJTNEWCHANNEL);
		addedNodes.add(astNew2);
		astLocal2.envIndex = RESERVED_ENV_IDX_RETVAL;
		astLocal2.TC_visited = true; // paranoid
		astLocal2.type = Types.CHANNEL;
		astLocal2.jjtAddChild(astNew2,0);
			
		// spawn f
		ASTSpawnStatement astSpawn1 = new ASTSpawnStatement(JJTSPAWNSTATEMENT);
		addedNodes.add(astSpawn1);
		astSpawn1.isMigration = false;
		astSpawn1.nameGiven = true;
		astSpawn1.CG_visited = false;
		astSpawn1.threadName = ((ASTRecursiveDefinitionCall)subchild).name;
		
		// spawn u.send(x)
		ASTSpawnStatement astSpawn2 = new ASTSpawnStatement(JJTSPAWNSTATEMENT);
		addedNodes.add(astSpawn2);
		astSpawn2.isMigration = false;
		astSpawn2.nameGiven = false;
		astSpawn2.CG_visited = false;
		ASTSend astSend = new ASTSend(JJTSEND);
		addedNodes.add(astSend);
		astSend.envIndex = RESERVED_ENV_IDX_PARAMS;
		Node param;
		for (int i=0; i<subchild.jjtGetNumChildren(); i++) {
			param = subchild.jjtGetChild(i);
			param.jjtSetParent(astSend);
			astSend.jjtAddChild(param,i);
		}
		astSpawn2.jjtAddChild(astSend,0);
		
		// create v.recv(y) node
		ASTRecv astRecv = new ASTRecv(JJTRECV);
		addedNodes.add(astRecv);
		astRecv.envIndex = RESERVED_ENV_IDX_RETVAL;
		astRecv.params = new Vector();
		if (!isVoid) {
			astRecv.params.add(
				new FormalParamDescriptor(((ASTLocalDeclarationStatement)child).envIndex));
		}
		
		// rebuild tree
		// now recDef node becomes the father of its encoding and if 
		// it's not void (= it's assigned to a variable) it takes place
		// of the local declaration node
		subchild.jjtAddChild(astLocal1,0);
		subchild.jjtAddChild(astLocal2,1);
		subchild.jjtAddChild(astSpawn1,2);
		subchild.jjtAddChild(astSpawn2,3);
		subchild.jjtAddChild(astRecv,4);
		if (!isVoid) {
			node.jjtAddChild(subchild, idx);
		}
		return 0;
	}

	public Object visit(ASTTypeDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}

	public Object visit(ASTImportDeclaration node, Object data) {
		// TODO Auto-generated method stub
		return null;
	}
}