/*
 * Created on 24-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package hipi.simulator;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Vector;

/**
 * 
 * @author Paolo Milazzo
 */
public class ChannelManager {
	
	// queue of threads waiting to react
	// they contain Threads 
	private ArrayList inq = new ArrayList(); // input queue
	private ArrayList outq = new ArrayList(); // output queue
	
	// reference to the deployment bag
	private ArrayList depBag;
	
	// reference to the dns
	private HashMap dns;
	
	// default constructor
	public ChannelManager(ArrayList depBagRef, HashMap dnsRef) {
		this.depBag = depBagRef;
		this.dns = dnsRef;
	}
	
	// reacts with an output or enqueue on input queue
	public void makeInput(SimThread thread, Vector message) {
		// if there are no outputs enqueues the thread
		if (outq.size()==0) {
			thread.setBlocked(true);
			inq.add(new Elem(thread, message));
		}
		// otherwise react with the first output in queue
		else {
			Elem popElem = (Elem) outq.remove(0);
			if (popElem.message.size()!=message.size()) {
				System.out.println("ARITY ERROR IN REACTION");
				System.out.println("BETWEEN " + thread.getName());
				System.out.println("AND" + popElem.thread.getName());
			}
			else {
				popElem.thread.setBlocked(false);
				depBag.add(popElem.thread);
				for (int i=0; i<message.size(); i++) {
					// if it's not <discard/>
					if (((Integer) message.get(i)).intValue() != -1) {
						thread.insertEnv(((Integer)message.get(i)).intValue(),
										popElem.message.get(i));
					}
				}
			}
			
		}
	}
	
	// reacts with an input or enqueue on output queue
	public void makeOutput(SimThread thread, Vector message) {
		// if there are no inputs enqueues the thread
		if (inq.size()==0) {
			thread.setBlocked(true);
			outq.add(new Elem(thread, message));
		}
		// otherwise react with the first input in queue
		else {
			Elem popElem = (Elem) inq.remove(0);
			if (popElem.isFwd) {
				VirtualMachine location = (VirtualMachine) dns.get(popElem.dest);
				if (location == null) {
					System.out.println("DESTINATION OF <fwd> DOESN'T EXIST IN THE NET");
					System.out.println("IN " + popElem.threadName);
				} 
				else {
					location.makeOutput(popElem.dest,thread,message);
				}
			} 
			else if (popElem.message.size() != message.size()) {
				System.out.println("ARITY ERROR IN REACTION");
				System.out.println("BETWEEN " + popElem.thread.getName());
				System.out.println("AND" + thread.getName());
			} 
			else {
				popElem.thread.setBlocked(false);
				depBag.add(popElem.thread);
				for (int i = 0; i < message.size(); i++) {
					// if it's not <discard/>
					if (((Integer) popElem.message.get(i)).intValue()!= -1) {
						popElem.thread.insertEnv(
							((Integer) popElem.message.get(i)).intValue(),
							message.get(i));
					}
				}
			}
		}
	}
	
	// reacts with an output or enqueue on input queue
	public void makeFwd(SimThread thread, URI dest) {
		// if there are no outputs enqueues the thread
		if (outq.size()==0) {
			thread.setBlocked(true);
			inq.add(new Elem(dest, thread.getName()));
		}
		// otherwise react with the first output in queue
		else {
			Elem popElem = (Elem) outq.remove(0);
			popElem.thread.setBlocked(false);
			VirtualMachine location = (VirtualMachine)dns.get(dest);
			if (location == null) {
				System.out.println("DESTINATION OF <fwd> DOESN'T EXIST IN THE NET");
				System.out.println("IN " + thread.getName());
			} 
			else {
				location.makeOutput(dest, popElem.thread, popElem.message);
			}
			
		}
	}
	
	// returns size of inq
	public int getInqSize() {
		return inq.size();
	}
	
	// return size of outq
	public int getOutqSize() {
		return outq.size();
	}
	

	// elements used into queues
	class Elem {
		// thread blocked
		public SimThread thread;
		
		// message sent (Integers, Strings and URIs)
		// or message received (Indexes of the environment)
		public Vector message;
		
		// Linear Forwarder flag
		public boolean isFwd;
		
		// Linear Forwarder destination
		public URI dest;
		
		// name of the thread
		public String threadName;
		
		// constructor for input and output
		public Elem(SimThread th, Vector msg) {
			this.thread = th;
			this.message = msg;
			this.isFwd = false;
			this.dest = null;
			this.threadName = th.getName();
		}
		
		// constructor for forwarder
		public Elem(URI destUri, String name) {
			this.thread = null;
			this.message = null;
			this.isFwd = true;
			this.dest = destUri;
			this.threadName = name;
		}
		
	}
}
