/*
 * Created on 24-lug-2003
 *
 * To change the template for this generated file go to
 * Window>Preferences>Java>Code Generation>Code and Comments
 */
package hipi.simulator;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * 
 * @author Paolo Milazzo
 */
public class VirtualMachine {

	// deployment bag: here running thread are stored
	// contains Threads
	private ArrayList depBag = new ArrayList();
	
	// channels owned by the machine
	// maps URIs into ChannelMagagers
	private HashMap channelSet = new HashMap();
	
	// reference to the dns
	private HashMap dns;
	
	// local channels counter (used for URI generation)
	private int chCounter = 0;
	
	// name of the virtual machine
	private String name;
	
	// default constructor
	public VirtualMachine(String vmName, HashMap dnsRef) {
		this.name = vmName;
		this.dns = dnsRef;
	}
	
	// creates a new channel and adds a reference into dns
	public void createChannel(URI uri) {
		ChannelManager cm = new ChannelManager(depBag,dns);
		channelSet.put(uri,cm);
		dns.put(uri,this);
	}
	
	// generates a new URI and creates a new local channel
	public URI createLocalChannel(String threadName) {
		URI uri;
		try {
			uri = new URI("ch://" + name + "/localchannels/" + threadName + "/idx_" + chCounter++);
			createChannel(uri);
			return uri;
		}
		catch (URISyntaxException e) {
			System.out.println("UNABLE TO CREATE LOCAL CHANNEL");
			System.out.println("URI = ch://" + name + "/localchannels/idx_" + chCounter);
		}
		return null;
	}
	
	// loads a new schedule
	public void loadSchedule(Document schedule) {
		// finds and executes the main thread
		NodeList threads = schedule.getElementsByTagName("thread");
		for (int i=0; i<threads.getLength(); i++) {
			String attribute = ((Element)threads.item(i)).getAttribute("name"); 
			if (attribute.equals("main")) {
				SimThread newThread = new SimThread(schedule, threads.item(i), this, null);
				depBag.add(newThread);
				System.out.println(name + ": New schedule loaded");
			}
		}
	}
	
	// executes one step
	public void step(boolean verbose) {
		if (depBag.size() !=0 ) {
			SimThread current = (SimThread) depBag.remove(0);
			current.step(verbose);
			if ((!current.isTerminated()) && (!current.isBlocked())) 
				depBag.add(current);
		}
	}
	
	// prints a view of the virtual machine
	public void view() {
		System.out.println(name + ":");
		// prints all channels
		System.out.println("   Channels - " + channelSet.size());
		Iterator chIter = channelSet.keySet().iterator();
		URI uri = null;
		while (chIter.hasNext()) {
			uri = (URI)chIter.next();
			System.out.println("      " + uri);
			// TODO print also waiting threads
			System.out.println("         INPUT QUEUE: " + 
							((ChannelManager)channelSet.get(uri)).getInqSize());
			System.out.println("         OUTPUT QUEUE:" + 
							((ChannelManager)channelSet.get(uri)).getOutqSize());
		}
		// prints all threads into the deployment bag
		System.out.println("   Active threads - " + depBag.size());
		// TODO print also active threads
	}

	// adds a thread into depBag
	public void addDepBag(SimThread newThread) {
		depBag.add(newThread);
	}	

	// enquiries the dns
	public VirtualMachine resolve(URI uri) {
		return (VirtualMachine) dns.get(uri);	
	}
	
	// execute an input on a channel
	public void makeInput(URI uri, SimThread thread, Vector message) {
		((ChannelManager) channelSet.get(uri)).makeInput(thread, message);
	}

	// execute an output on a channel
	public void makeOutput(URI uri, SimThread thread, Vector message) {
		((ChannelManager) channelSet.get(uri)).makeOutput(thread, message);
	}
	
	// execute a forwarding on a channel
	public void makeFwd(URI uri, SimThread thread, URI dest) {
		((ChannelManager) channelSet.get(uri)).makeFwd(thread,dest);
	}
	
	// returns name
	public String getName() {
		return name;
	}
}
